@@ -9,14 +9,12 @@ import (
 	"net/url"
 	"os"
 	"os/signal"
-	"path"
 	"path/filepath"
 	"regexp"
 	"runtime"
 	"runtime/pprof"
 	"sort"
 	"strings"
-	"syscall"
 	"time"
 
 	"github.com/djherbis/atime"
@@ -75,12 +73,15 @@ type Task struct {
 // NewTask returns a new Task.
 func NewTask(root, input, output string, sync bool) (Task, error) {
 	t := Task{[]string{input}, output, sync}
-	if 0 < len(output) && output[len(output)-1] == '/' {
+	if IsDir(output) {
+		root = filepath.Dir(root)
+	}
+	if 0 < len(output) && output[len(output)-1] == os.PathSeparator {
 		rel, err := filepath.Rel(root, input)
 		if err != nil {
 			return Task{}, err
 		}
-		t.dst = path.Join(output, path.Base(root), filepath.ToSlash(rel))
+		t.dst = filepath.Join(output, rel)
 	}
 	return t, nil
 }
@@ -149,6 +150,7 @@ func run() int {
 	flag.BoolVar(&htmlMinifier.KeepQuotes, "html-keep-quotes", false, "Preserve quotes around attribute values")
 	flag.IntVar(&jsMinifier.Precision, "js-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
 	flag.BoolVar(&jsMinifier.KeepVarNames, "js-keep-var-names", false, "Preserve original variable names")
+	flag.BoolVar(&jsMinifier.NoNullishOperator, "js-no-nullish-operator", false, "Don't use the ?? nullish coalescing operator")
 	flag.IntVar(&jsonMinifier.Precision, "json-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
 	flag.BoolVar(&jsonMinifier.KeepNumbers, "json-keep-numbers", false, "Preserve original numbers instead of minifying them")
 	flag.IntVar(&svgMinifier.Precision, "svg-precision", 0, "Number of significant digits to preserve in numbers, 0 is all")
@@ -301,22 +303,31 @@ func run() int {
 	////////////////
 
 	for i, input := range inputs {
-		inputs[i] = sanitizePath(input)
+		inputs[i] = filepath.Clean(input)
 	}
 
 	// set output, empty means stdout, ending in slash means a directory, otherwise a file
 	dirDst := false
 	if output != "" {
-		dirSrc := 1 < len(inputs) || len(inputs) == 1 && inputs[0][len(inputs[0])-1] == '/'
-		if !bundle && dirSrc && output[len(output)-1] != '/' {
-			output += "/"
+		dirDst = IsDir(output)
+		if !dirDst {
+			if 1 < len(inputs) && !bundle {
+				dirDst = true
+			} else if len(inputs) == 1 {
+				if info, err := os.Lstat(inputs[0]); err == nil && !bundle && info.Mode().IsDir() && info.Mode()&os.ModeSymlink == 0 {
+					dirDst = true
+				}
+			}
 		}
-		output = sanitizePath(output)
-		dirDst = output[len(output)-1] == '/'
 		if dirDst && bundle {
 			Error.Println("--bundle requires destination to be stdout or a file")
 			return 1
 		}
+
+		output = filepath.Clean(output)
+		if dirDst {
+			output += string(os.PathSeparator)
+		}
 	} else if 1 < len(inputs) {
 		Error.Println("must specify --bundle for multiple input files with stdout destination")
 		return 1
@@ -335,12 +346,6 @@ func run() int {
 			Info.Println("minify from stdin")
 		}
 	}
-	if dirDst {
-		if err := os.MkdirAll(output, 0777); err != nil {
-			Error.Println(err)
-			return 1
-		}
-	}
 
 	var tasks []Task
 	var roots []string
@@ -369,6 +374,14 @@ func run() int {
 		tasks = tasks[:1]
 	}
 
+	// make output directory
+	if dirDst {
+		if err := os.MkdirAll(output, 0777); err != nil {
+			Error.Println(err)
+			return 1
+		}
+	}
+
 	////////////////
 
 	m = min.New()
@@ -384,101 +397,102 @@ func run() int {
 		return 1
 	}
 
-	numWorkers := 1
-	if !verbose && len(tasks) > 1 {
-		numWorkers = 4
-		if n := runtime.NumCPU(); n > numWorkers {
-			numWorkers = n
-		}
-	}
-
+	fails := 0
 	start := time.Now()
-
-	chanTasks := make(chan Task, 100)
-	chanFails := make(chan int, numWorkers)
-	for n := 0; n < numWorkers; n++ {
-		go minifyWorker(chanTasks, chanFails)
-	}
-
-	if !watch {
+	if !watch && (len(tasks) == 1 || verbose) {
 		for _, task := range tasks {
-			chanTasks <- task
+			if ok := minify(task); !ok {
+				fails++
+			}
 		}
 	} else {
-		watcher, err := NewWatcher(recursive)
-		if err != nil {
-			Error.Println(err)
-			return 1
+		numWorkers := runtime.NumCPU()
+		if verbose {
+			numWorkers = 1
+		} else if numWorkers < 4 {
+			numWorkers = 4
 		}
-		defer watcher.Close()
 
-		changes := watcher.Run()
+		chanTasks := make(chan Task, 20)
+		chanFails := make(chan int, numWorkers)
+		for n := 0; n < numWorkers; n++ {
+			go minifyWorker(chanTasks, chanFails)
+		}
 		for _, task := range tasks {
 			chanTasks <- task
 		}
 
-		autoDir := false
-		files := roots
-		if !recursive {
-			files = inputs
-		}
-		for _, filename := range files {
-			watcher.AddPath(filename)
-			if filename == output {
-				autoDir = true
+		if watch {
+			watcher, err := NewWatcher(recursive)
+			if err != nil {
+				Error.Println(err)
+				return 1
 			}
-		}
-		skip := map[string]bool{}
+			defer watcher.Close()
 
-		c := make(chan os.Signal, 1)
-		signal.Notify(c, os.Interrupt)
-		for changes != nil {
-			select {
-			case <-c:
-				watcher.Close()
-			case file, ok := <-changes:
-				if !ok {
-					changes = nil
-					break
+			changes := watcher.Run()
+			autoDir := false
+			files := roots
+			if !recursive {
+				files = inputs
+			}
+			for _, filename := range files {
+				watcher.AddPath(filename)
+				if filename == output {
+					autoDir = true
 				}
-				file = sanitizePath(file)
+			}
+			skip := map[string]bool{}
 
-				// find longest common path among roots
-				root := ""
-				for _, path := range roots {
-					pathRel, err1 := filepath.Rel(path, file)
-					rootRel, err2 := filepath.Rel(root, file)
-					if err2 != nil || err1 == nil && len(pathRel) < len(rootRel) {
-						root = path
-					}
-				}
-
-				if autoDir && root == output {
-					// skip files in output directory (which is also an input directory) for the first change
-					// skips files that are not minified and stay put as they are not explicitly copied, but that's ok
-					if _, ok := skip[file]; !ok {
-						skip[file] = true
+			c := make(chan os.Signal, 1)
+			signal.Notify(c, os.Interrupt)
+			for changes != nil {
+				select {
+				case <-c:
+					watcher.Close()
+				case file, ok := <-changes:
+					if !ok {
+						changes = nil
 						break
 					}
-				}
+					file = filepath.Clean(file)
 
-				if !verbose {
-					Info.Println(file, "changed")
+					// find longest common path among roots
+					root := ""
+					for _, path := range roots {
+						pathRel, err1 := filepath.Rel(path, file)
+						rootRel, err2 := filepath.Rel(root, file)
+						if err2 != nil || err1 == nil && len(pathRel) < len(rootRel) {
+							root = path
+						}
+					}
+
+					if autoDir && root == output {
+						// skip files in output directory (which is also an input directory) for the first change
+						// skips files that are not minified and stay put as they are not explicitly copied, but that's ok
+						if _, ok := skip[file]; !ok {
+							skip[file] = true
+							break
+						}
+					}
+
+					if !verbose {
+						Info.Println(file, "changed")
+					}
+					task, err := NewTask(root, file, output, !fileMatches(file))
+					if err != nil {
+						Error.Println(err)
+						return 1
+					}
+					chanTasks <- task
 				}
-				task, err := NewTask(root, file, output, !fileMatches(file))
-				if err != nil {
-					Error.Println(err)
-					return 1
-				}
-				chanTasks <- task
 			}
 		}
-	}
 
-	fails := 0
-	close(chanTasks)
-	for n := 0; n < numWorkers; n++ {
-		fails += <-chanFails
+		close(chanTasks)
+		for n := 0; n < numWorkers; n++ {
+			fails += <-chanFails
+		}
 	}
 
 	if verbose && !watch {
@@ -500,25 +514,6 @@ func minifyWorker(chanTasks <-chan Task, chanFails chan<- int) {
 	chanFails <- fails
 }
 
-func sanitizePath(p string) string {
-	p = filepath.ToSlash(p) // replace \ with / for Windows
-	isCur := path.Base(p) == "."
-	isDir := p[len(p)-1] == '/' || isCur
-	p = path.Clean(p)
-	if p == "." {
-		return "./"
-	}
-	if isDir {
-		p += "/"
-	} else if info, err := os.Lstat(p); err == nil && info.Mode().IsDir() && info.Mode()&os.ModeSymlink == 0 {
-		p += "/"
-	}
-	if isCur {
-		p += "./"
-	}
-	return p
-}
-
 func fileMatches(filename string) bool {
 	if pattern != nil && !pattern.MatchString(filename) {
 		return false
@@ -526,7 +521,7 @@ func fileMatches(filename string) bool {
 		return true
 	}
 
-	ext := path.Ext(filename)
+	ext := filepath.Ext(filename)
 	if 0 < len(ext) {
 		ext = ext[1:]
 	}
@@ -536,18 +531,6 @@ func fileMatches(filename string) bool {
 	return true
 }
 
-//type DirEntry struct {
-//	fs.FileInfo
-//}
-//
-//func (d DirEntry) Type() fs.FileMode {
-//	return d.Mode().Type()
-//}
-//
-//func (d DirEntry) Info() (fs.FileInfo, error) {
-//	return d.FileInfo, nil
-//}
-
 func createTasks(inputs []string, output string) ([]Task, []string, error) {
 	tasks := []Task{}
 	roots := []string{}
@@ -563,12 +546,13 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 			return nil, nil, err
 		}
 
+		root := filepath.Dir(input) + string(os.PathSeparator) + "."
 		if info.Mode()&os.ModeSymlink != 0 {
 			if !sync {
 				Warning.Println("--sync not specified, omitting symbolic link", input)
 				continue
 			}
-			task, err := NewTask(path.Dir(input), input, output, true)
+			task, err := NewTask(root, input, output, true)
 			if err != nil {
 				return nil, nil, err
 			}
@@ -576,7 +560,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 		} else if info.Mode().IsRegular() {
 			valid := pattern == nil || pattern.MatchString(info.Name())
 			if valid || sync {
-				task, err := NewTask(path.Dir(input), input, output, !valid)
+				task, err := NewTask(root, input, output, !valid)
 				if err != nil {
 					return nil, nil, err
 				}
@@ -587,7 +571,8 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 				Warning.Println("--recursive not specified, omitting directory", input)
 				continue
 			}
-			roots = append(roots, input)
+			root = input
+			roots = append(roots, root)
 
 			var walkFn func(string, DirEntry, error) error
 			walkFn = func(path string, d DirEntry, err error) error {
@@ -601,16 +586,24 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 					}
 					return nil
 				}
-				path = sanitizePath(path)
+
+				input := path
+				if filepath.IsAbs(input) {
+					// path is absolute, clean it
+					input = filepath.Clean(input)
+				} else {
+					// path is relative, make an absolute path - Join() returns a clean result
+					input = filepath.Join(root, input)
+				}
 
 				if !preserveLinks && d.Type()&os.ModeSymlink != 0 {
 					// follow and dereference symlinks
-					info, err := os.Stat(path)
+					info, err := os.Stat(input)
 					if err != nil {
 						return err
 					}
 					if info.IsDir() {
-						return WalkDir(DirFS(input), path, walkFn)
+						return WalkDir(DirFS(root), path, walkFn)
 					}
 					d = &statDirEntry{info}
 				}
@@ -621,7 +614,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 						Warning.Println("--sync not specified, omitting symbolic link", path)
 						return nil
 					}
-					task, err := NewTask(input, path, output, true)
+					task, err := NewTask(root, input, output, true)
 					if err != nil {
 						return err
 					}
@@ -629,7 +622,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 				} else if d.Type().IsRegular() {
 					valid := fileMatches(d.Name())
 					if valid || sync {
-						task, err := NewTask(input, path, output, !valid)
+						task, err := NewTask(root, input, output, !valid)
 						if err != nil {
 							return err
 						}
@@ -638,7 +631,7 @@ func createTasks(inputs []string, output string) ([]Task, []string, error) {
 				}
 				return nil
 			}
-			if err := WalkDir(DirFS("."), path.Clean(input), walkFn); err != nil {
+			if err := WalkDir(DirFS(root), ".", walkFn); err != nil {
 				return nil, nil, err
 			}
 		} else {
@@ -658,8 +651,9 @@ func openInputFile(input string) (io.ReadCloser, error) {
 			r, ferr = os.Open(input)
 			return attempt < 5, ferr
 		})
+
 		if err != nil {
-			return nil, err
+			return nil, fmt.Errorf("open input file %q: %w", input, err)
 		}
 	}
 	return r, nil
@@ -670,16 +664,19 @@ func openOutputFile(output string) (*os.File, error) {
 	if output == "" {
 		w = os.Stdout
 	} else {
-		if err := os.MkdirAll(path.Dir(output), 0777); err != nil {
-			return nil, err
+		dir := filepath.Dir(output)
+		if err := os.MkdirAll(dir, 0777); err != nil {
+			return nil, fmt.Errorf("creating directory %q: %w", dir, err)
 		}
+
 		err := try.Do(func(attempt int) (bool, error) {
 			var ferr error
 			w, ferr = os.OpenFile(output, os.O_WRONLY|os.O_TRUNC|os.O_CREATE, 0666)
 			return attempt < 5, ferr
 		})
+
 		if err != nil {
-			return nil, err
+			return nil, fmt.Errorf("open output file %q: %w", output, err)
 		}
 	}
 	return w, nil
@@ -691,7 +688,7 @@ func createSymlink(input, output string) error {
 			return err
 		}
 	}
-	if err := os.MkdirAll(path.Dir(output), 0777); err != nil {
+	if err := os.MkdirAll(filepath.Dir(output), 0777); err != nil {
 		return err
 	}
 	if err := os.Symlink(input, output); err != nil {
@@ -722,7 +719,7 @@ func minify(t Task) bool {
 	fileMimetype := mimetype
 	if mimetype == "" && !t.sync {
 		for _, src := range t.srcs {
-			ext := path.Ext(src)
+			ext := filepath.Ext(src)
 			if 0 < len(ext) {
 				ext = ext[1:]
 			}
@@ -753,7 +750,7 @@ func minify(t Task) bool {
 	} else {
 		// rename original when overwriting
 		for i := range t.srcs {
-			if t.srcs[i] == t.dst {
+			if sameFile, _ := SameFile(t.srcs[i], t.dst); sameFile {
 				t.srcs[i] += ".bak"
 				err := try.Do(func(attempt int) (bool, error) {
 					ferr := os.Rename(t.dst, t.srcs[i])
@@ -768,15 +765,23 @@ func minify(t Task) bool {
 		}
 	}
 
-	fr, err := newConcatFileReader(t.srcs, openInputFile)
+	var err error
+	var fr io.ReadCloser
+	var fw io.WriteCloser
+	if len(t.srcs) == 1 {
+		fr, err = openInputFile(t.srcs[0])
+	} else {
+		fr, err = newConcatFileReader(t.srcs, openInputFile)
+		if err == nil && fileMimetype == filetypeMime["js"] {
+			fr.(*concatFileReader).SetSeparator([]byte("\n"))
+		}
+	}
 	if err != nil {
 		Error.Println(err)
 		return false
 	}
-	if fileMimetype == filetypeMime["js"] {
-		fr.SetSeparator([]byte("\n"))
-	}
-	fw, err := openOutputFile(t.dst)
+
+	fw, err = openOutputFile(t.dst)
 	if err != nil {
 		Error.Println(err)
 		fr.Close()
@@ -798,22 +803,25 @@ func minify(t Task) bool {
 	}
 
 	r := newCountingReader(fr)
-	var w *countingWriter
-	if fw == os.Stdout {
-		w = newCountingWriter(fw)
-	} else {
-		w = newCountingWriter(bufio.NewWriter(fw))
-	}
+	w := newCountingWriter(fw)
+	bw := bufio.NewWriter(w)
 
 	success := true
 	startTime := time.Now()
-	if err = m.Minify(fileMimetype, w, r); err != nil {
+	if err = m.Minify(fileMimetype, bw, r); err != nil {
+		bw.Flush()
 		if t.dst == "" && 0 < w.N {
 			fmt.Fprintf(os.Stdout, "\n\n")
 		}
 		Error.Println("cannot minify "+srcName+":", err)
 		success = false
+	} else {
+		bw.Flush()
 	}
+
+	fr.Close()
+	fw.Close()
+
 	if verbose {
 		dur := time.Since(startTime)
 		speed := "Inf MB"
@@ -825,7 +833,7 @@ func minify(t Task) bool {
 			ratio = float64(w.N) / float64(r.N)
 		}
 
-		stats := fmt.Sprintf("(%9v, %6v, %6v, %5.1f%%, %6v/s)", dur, humanize.Bytes(uint64(r.N)), humanize.Bytes(uint64(w.N)), ratio*100, speed)
+		stats := fmt.Sprintf("(%9v, %6v, %6v, %5.1f%%, %6v/s)", dur, humanize.Bytes(r.N), humanize.Bytes(w.N), ratio*100, speed)
 		if srcName != dstName {
 			Info.Println(stats, "-", srcName, "to", dstName)
 		} else {
@@ -833,12 +841,6 @@ func minify(t Task) bool {
 		}
 	}
 
-	fr.Close()
-	if bw, ok := w.Writer.(*bufio.Writer); ok {
-		bw.Flush()
-	}
-	fw.Close()
-
 	// remove original that was renamed, when overwriting files
 	for i := range t.srcs {
 		if t.srcs[i] == t.dst+".bak" {
@@ -879,8 +881,8 @@ func preserveAttributes(src, dst string) {
 			}
 		}
 		if preserveOwnership {
-			if stat_t, ok := srcInfo.Sys().(*syscall.Stat_t); ok {
-				err = os.Chown(dst, int(stat_t.Uid), int(stat_t.Gid))
+			if uid, gid, ok := getOwnership(srcInfo); ok {
+				err = os.Chown(dst, uid, gid)
 				if err != nil {
 					Warning.Println(err)
 				}
@@ -889,7 +891,6 @@ func preserveAttributes(src, dst string) {
 			}
 		}
 		if preserveTimestamps {
-			fmt.Println(atime.Get(srcInfo), srcInfo.ModTime())
 			err = os.Chtimes(dst, atime.Get(srcInfo), srcInfo.ModTime())
 			if err != nil {
 				Warning.Println(err)
@@ -897,3 +898,25 @@ func preserveAttributes(src, dst string) {
 		}
 	}
 }
+
+// IsDir returns true if the passed string looks like it specifies a directory, false otherwise.
+func IsDir(dir string) bool {
+	info, err := os.Lstat(dir)
+	return err == nil && info.Mode().IsDir() && info.Mode()&os.ModeSymlink == 0
+}
+
+// SameFile returns true if the two file paths specify the same path.
+// While Linux is case-preserving case-sensitive (and therefore a string comparison will work),
+// Windows is case-preserving case-insensitive; we use os.SameFile() to work cross-platform.
+func SameFile(filename1 string, filename2 string) (bool, error) {
+	fi1, err := os.Stat(filename1)
+	if err != nil {
+		return false, err
+	}
+
+	fi2, err := os.Stat(filename2)
+	if err != nil {
+		return false, err
+	}
+	return os.SameFile(fi1, fi2), nil
+}
